// LapseCam v0.02
//
// by Rick Murray
// 1st April 2023
//
//
// The ESP32-Cam will wake up every minute, retrieve a sequential number, and then
// take a photograph with that number. Once done, it'll go to sleep until the next
// minute.
// This is intended for a cheap ESP32-CAM board with an OV2640 (2mpix) camera
// module. If you have a different camera, you may need to tweak a few things.
//
// If there is a file on the SD card called "reset.txt", then the counter will be
// reset, and the text file then deleted.
//
// If there is a file on the SD card called "config.txt", then this can be used
// to change some of the behaviour.
//   f=<num>    Frame size (0-10)         [default = 10 = UXGA]
//   q=<num>    Quality setting (1-63)    [default = 5]
//   i=<num>    Interval (15+ seconds++)  [default = 60 = 1 minute]
//   (you only need include those lines you change)
//



// Basics
#include "Arduino.h"
// The camera
#include "esp_camera.h"
// The filesystem
#include "FS.h"
#include "SD_MMC.h"
// Turn off the brownout detection (stops booting)
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"
// Control the RTC for our wakeup
#include "driver/rtc_io.h"
// Access the internal flash
#include <EEPROM.h>
#define EEPROM_SIZE 2



// These pins are defined for the cheap ESP32-CAM
// board that you can find all over the place.
// They're all clones of the AI Thinker board.

#define RESET_GPIO_NUM    -1
#define XCLK_GPIO_NUM      0
#define FLASHLIGHT_PIN     4 // doubled with SD card
#define Y2_GPIO_NUM        5
#define Y3_GPIO_NUM       18
#define Y4_GPIO_NUM       19
#define Y5_GPIO_NUM       21
#define PCLK_GPIO_NUM     22
#define HREF_GPIO_NUM     23
#define VSYNC_GPIO_NUM    25
#define SIOD_GPIO_NUM     26
#define SIOC_GPIO_NUM     27
#define PWDN_GPIO_NUM     32
#define LED_PIN           33
#define Y8_GPIO_NUM       34
#define Y9_GPIO_NUM       35
#define Y6_GPIO_NUM       36
#define Y7_GPIO_NUM       39

// LED is inverted, so make some easy defines
#define LED_ON           LOW
#define LED_OFF         HIGH



RTC_DATA_ATTR int bootcount = 0;



void setup()
{
   int      counter = 0;
   char     filename[32] = "";
   int64_t  ticks = esp_timer_get_time();
   uint64_t interval = 60000000;        // default 1 minute
   uint8_t  sdtype = CARD_NONE;
   uint64_t sdsize = 0;
   uint64_t sdfree = 0;
   int      quality = 5;                // default
   int      framesize = FRAMESIZE_UXGA; // default
   char     confbuff[8] = "";
   int      confread = 0;
   int      confval = 0;
   int      byteswritten = 0;


   // First thing, kill the brownout detection
   WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);


   // Say hello
   Serial.begin(115200);
   Serial.println();
   Serial.println("Rick's LapseCam starting.");
   bootcount++;
   Serial.printf("This is boot %d since powered up.\n", bootcount);

   // Turn on the status LED as well
   pinMode(LED_PIN, OUTPUT);
   digitalWrite(LED_PIN, LED_ON);


   // Set up the camera
   camera_config_t config;
   config.ledc_channel = LEDC_CHANNEL_0;
   config.ledc_timer = LEDC_TIMER_0;
   config.pin_d0 = Y2_GPIO_NUM;
   config.pin_d1 = Y3_GPIO_NUM;
   config.pin_d2 = Y4_GPIO_NUM;
   config.pin_d3 = Y5_GPIO_NUM;
   config.pin_d4 = Y6_GPIO_NUM;
   config.pin_d5 = Y7_GPIO_NUM;
   config.pin_d6 = Y8_GPIO_NUM;
   config.pin_d7 = Y9_GPIO_NUM;
   config.pin_xclk = XCLK_GPIO_NUM;
   config.pin_pclk = PCLK_GPIO_NUM;
   config.pin_vsync = VSYNC_GPIO_NUM;
   config.pin_href = HREF_GPIO_NUM;
   config.pin_sscb_sda = SIOD_GPIO_NUM;
   config.pin_sscb_scl = SIOC_GPIO_NUM;
   config.pin_pwdn = PWDN_GPIO_NUM;
   config.pin_reset = RESET_GPIO_NUM;
   config.xclk_freq_hz = 20000000;
   config.pixel_format = PIXFORMAT_JPEG;

   // I don't check if PSRAM found, I know my device has it.

   // Initialise with maximum memory for images
   Serial.println("Starting camera.");
   config.frame_size = FRAMESIZE_UXGA;
   config.jpeg_quality = 2;
   config.fb_count = 2; // run continual
   esp_err_t err = esp_camera_init(&config);
   if (err != ESP_OK)
   {
      Serial.printf("ERROR: Camera start failed with error &%x.\n", err);
      for (counter = 1; counter < 20; counter++)
      {
         // Blink LED so we know it failed
         digitalWrite(LED_PIN, LED_ON);
         delay(100);
         digitalWrite(LED_PIN, LED_OFF);
         delay(100);
      }
      ESP.restart(); // then hit the big red button
      return;
   }


   // Read our EEPROM for the 16 bit sequential counter
   // 16 bits, files image00000.jpg -> image65535.jpg
   EEPROM.begin(EEPROM_SIZE);
   counter = EEPROM.read(0) + ( EEPROM.read(1) << 8 );
   if ( counter == 0xFFFF )
      counter = 0; // initial state is both bytes &FF
   Serial.printf("Counter for this image is %d.\n", counter);


   // As we held the flash LED off when powered down, it must be
   // released or else the SD card won't work.
   rtc_gpio_hold_dis(GPIO_NUM_4);


   // Get the SD card running
   Serial.println("Starting SD card.");
   if ( !SD_MMC.begin() )
   {
      Serial.println("ERROR: SD Card could not be started.");
      for (counter = 1; counter < 40; counter++)
      {
         // Blink LED faster so we know it failed
         digitalWrite(LED_PIN, LED_ON);
         delay(50);
         digitalWrite(LED_PIN, LED_OFF);
         delay(50);
      }
      ESP.restart(); // then hit the big red button
   }

   // Check there's something IN the SD slot!
   sdtype = SD_MMC.cardType();
   if (sdtype == CARD_NONE)
   {
      Serial.println("ERROR: No SD card! Halting.");
      while (true)
      {
         // Just loop, this isn't a glitch, it's a show stopper
         digitalWrite(LED_PIN, LED_ON);
         delay(50);
         digitalWrite(LED_PIN, LED_OFF);
         delay(50);
      }
   }

   sdsize = SD_MMC.cardSize() / (1024 * 1024);
   sdfree = (SD_MMC.totalBytes() - SD_MMC.usedBytes()) / (1024 * 1024);
   Serial.printf("SD card: capacity %lluMB, free %lluMB\n", sdsize, sdfree);

   // Associate a filing system
   fs::FS &fs = SD_MMC;


   // Is the "reset.txt" file present?
   if ( fs.exists("/reset.txt") )
   {
      Serial.println("The reset.txt file is present, resetting counter.");
      counter = 0;
      fs.remove("/reset.txt");
   }


   // Is the "config.txt" file present?
   if ( fs.exists("/config.txt") )
   {
      Serial.println("The config.txt file is present, loading settings.");

      File conffile = fs.open("/config.txt", FILE_READ);
      if ( !conffile )
      {
         Serial.println("Failed to open config.txt, using defaults.");
      }
      else
      {
         while ( conffile.available() )
         {
            confread = conffile.readBytesUntil('\n', confbuff, 8);

            if ( confread > 0 )
            {
               // f=<num> to set framesize
               if ( tolower(confbuff[0]) == 'f' )
               {
                  confval = atoi(confbuff + 2);
                  if ( ( confval > 0 ) && ( confval < 11 ) )
                  {
                     // This relates to the ORIGINAL resolutions
                     // list. The use of the defines is because
                     // the list doesn't match what's there now.
                     Serial.print("-> Setting framesize to ");
                     switch (confval)
                     {
                        case  0 : // fall through
                        case  1 : Serial.println("QQVGA (160x120).");
                                  confval = FRAMESIZE_QQVGA;
                                  // 1 = QQVGA2 (128x160), this was
                                  //     an odd size, not present in
                                  //     current builds.
                                  break;
                        case  2 : Serial.println("QCIF (176x144).");
                                  confval = FRAMESIZE_QCIF;
                                  break;
                        case  3 : Serial.println("HQVGA (240x176).");
                                  confval = FRAMESIZE_HQVGA;
                                  break;
                        case  4 : Serial.println("QVGA (320x240).");
                                  confval = FRAMESIZE_QVGA;
                                  break;
                        case  5 : Serial.println("CIF (400x296).");
                                  confval = FRAMESIZE_CIF;
                                  break;
                        case  6 : Serial.println("VGA (640x480).");
                                  confval = FRAMESIZE_VGA;
                                  break;
                        case  7 : Serial.println("SVGA (800x600).");
                                  confval = FRAMESIZE_SVGA;
                                  break;
                        case  8 : Serial.println("XGA (1024x768).");
                                  confval = FRAMESIZE_XGA;
                                  break;
                        case  9 : Serial.println("SXGA (1280x1024).");
                                  confval = FRAMESIZE_SXGA;
                                  break;
                        case 10 : Serial.println("UXGA (1600x1200).");
                                  confval = FRAMESIZE_UXGA;
                                  break;
                        // We do not support 96x96 and HD (1280x720)
                        // as these are newer than my SDK.
                     }
                     framesize = confval;
                  }
                  else
                  {
                     Serial.printf("Invalid framesize %d.\n", confval);
                  }
               }

               // q=<num> to set quality
               if ( tolower(confbuff[0]) == 'q' )
               {
                  confval = atoi(confbuff + 2);
                  if ( ( confval > 0 ) && ( confval < 64 ) )
                  {
                     Serial.printf("-> Setting quality to %d.\n", confval);
                     quality = confval;
                  }
                  else
                  {
                     Serial.printf("Invalid quality %d.\n", confval);
                  }
               }

               // i=<num> to set timer interval
               if ( tolower(confbuff[0]) == 'i' )
               {
                  confval = atoi(confbuff + 2);
                  if ( confval < 15 )
                  {
                     Serial.printf("Interval must be at least 15 seconds.\n");
                     confval = 15;
                  }
                  Serial.printf("-> Setting timer interval to %d seconds.\n", confval);

                  interval = (uint64_t)confval * (uint64_t)1000000;
               }
            } // if (confread > 0 )
         } // while ( fs.available() )

         conffile.close();
      } // if config file opened
   } // if config file exists


   // Set the camera options
   sensor_t * s = esp_camera_sensor_get();
   s->set_framesize(s, (framesize_t)framesize);
   s->set_quality(s, quality);
   // other options can just be the camera defaults


   // Now wait five seconds for the AEC to settle
   Serial.println("Waiting five seconds for AEC etc to settle.");
   delay(5000);


   // Take a picture
   Serial.println("Smile! :) Taking picture.");
   camera_fb_t * fb = NULL;
   fb = esp_camera_fb_get();
   if (!fb)
   {
      Serial.println("ERROR: Failed to take picture!");
      ESP.restart(); // what else can we do?
      return;
   }


   // Create subdirectory if necessary
   sprintf(filename, "/%05d", counter);
   if ( ( filename[4] == '0' ) && ( filename[5] == '0' ) )
      if ( ! fs.exists(filename) )
         fs.mkdir(filename); // make the subdirectory if it doesn't exist

   // Now make a filename
   sprintf(filename, "/%05d/image%05d.jpg", counter, counter);
   filename[4] = '0'; // grotty but
   filename[5] = '0'; // quick ;)
   Serial.printf("Writing picture as \"%s\".\n", filename);


   // Open the output file and write to it
   File file = fs.open(filename, FILE_WRITE);
   if ( !file )
   {
      Serial.println("ERROR: Failed to open file for writing.");
      ESP.restart();
      // ##TODO## Trap if this is because the media is full!
   }
   else
   {
      byteswritten = file.write(fb->buf,     fb->len);
      file.close();          // buffer data, buffer length

      Serial.printf("Saved %d of %d bytes, updating counter.\n",
                    byteswritten, fb->len);

      // Trap if data couldn't be written (usually means disc full)
      if ( byteswritten != fb->len )
      {
         Serial.println("ERROR: Unable to write all data - disc full?");
         while (true)
         {
            // Just loop, slow LED flash, it's a show stopper
            digitalWrite(LED_PIN, LED_ON);
            delay(500);
            digitalWrite(LED_PIN, LED_OFF);
            delay(500);
         }
      }

      // If we're here, file saved correctly, so update
      // the sequential counter and commit it to EEPROM.
      counter++;
      EEPROM.write(0,  (counter       & 255));
      EEPROM.write(1, ((counter >> 8) & 255));
      EEPROM.commit();
   }


   // Tidy up
   esp_camera_fb_return(fb);
   SD_MMC.end();

   // Turn off the flash LED and force it off
   pinMode(FLASHLIGHT_PIN, OUTPUT);
   digitalWrite(FLASHLIGHT_PIN, LOW);
   rtc_gpio_hold_en(GPIO_NUM_4);

   // Tell the RTC timer to prod us when the time has elapsed
   ticks = esp_timer_get_time() - ticks; // our elapsed time
   ticks = interval - ticks; // how much remains of our interval time
   esp_sleep_enable_timer_wakeup(ticks);


   // Final status
   Serial.println("Going into deep sleep for %llu seconds, seeya!", (ticks / 1000000));
   delay(500);
   Serial.flush();

   // Now turn off the status LED as we're done
   pinMode(LED_PIN, OUTPUT);
   digitalWrite(LED_PIN, LED_OFF);


   // Now do it.
   esp_deep_sleep_start();


   // we never come here
}



void loop()
{
   // isn't used
}
